from .inference import (
    NSFWImageDetectionInput,
    NSFWImageDetectionOutput,
    run_nsfw_estimation,
)

__all__ = [
    "NSFWImageDetectionInput",
    "NSFWImageDetectionOutput",
    "run_nsfw_estimation",
]



# Originally from https://github.com/httpie/cli/blob/master/httpie/cli/nested_json/interpret.py
#
# Copyright © 2012-2022 Jakub Roztocil <jakub@roztocil.co>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of the copyright holder nor the names of its contributors
#        may be used to endorse or promote products derived from this software
#        without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

from collections.abc import Iterable, Iterator
from enum import Enum, auto
from typing import Any, NamedTuple, Union

EMPTY_STRING = ""
HIGHLIGHTER = "^"
OPEN_BRACKET = "["
CLOSE_BRACKET = "]"
BACKSLASH = "\\"


class TokenKind(Enum):
    TEXT = auto()
    NUMBER = auto()
    LEFT_BRACKET = auto()
    RIGHT_BRACKET = auto()
    PSEUDO = auto()  # Not a real token, use when representing location only.

    def to_name(self) -> str:
        for key, value in OPERATORS.items():
            if value is self:
                return repr(key)
        return "a " + self.name.lower()


OPERATORS = {
    OPEN_BRACKET: TokenKind.LEFT_BRACKET,
    CLOSE_BRACKET: TokenKind.RIGHT_BRACKET,
}
SPECIAL_CHARS = OPERATORS.keys() | {BACKSLASH}
LITERAL_TOKENS = [
    TokenKind.TEXT,
    TokenKind.NUMBER,
]


class Token(NamedTuple):
    kind: TokenKind
    value: str | int
    start: int
    end: int


class PathAction(Enum):
    KEY = auto()
    INDEX = auto()
    APPEND = auto()
    # Pseudo action, used by the interpreter
    SET = auto()

    def to_string(self) -> str:
        return self.name.lower()


class Path:
    def __init__(
        self,
        kind: PathAction,
        accessor: str | int | None = None,
        tokens: list[Token] | None = None,
        is_root: bool = False,
    ):
        self.kind = kind
        self.accessor = accessor
        self.tokens = tokens or []
        self.is_root = is_root

    def reconstruct(self) -> str:
        if self.kind is PathAction.KEY:
            if self.is_root:
                return str(self.accessor)
            return OPEN_BRACKET + str(self.accessor) + CLOSE_BRACKET
        elif self.kind is PathAction.INDEX:
            return OPEN_BRACKET + str(self.accessor) + CLOSE_BRACKET
        elif self.kind is PathAction.APPEND:
            return OPEN_BRACKET + CLOSE_BRACKET
        else:
            raise ValueError(f"Unexpected path action: {self.kind}")


class NestedJSONArray(list):
    """Denotes a top-level JSON array."""


class NestedJSONSyntaxError(ValueError):
    def __init__(
        self,
        source: str,
        token: Token | None,
        message: str,
        message_kind: str = "Syntax",
    ) -> None:
        self.source = source
        self.token = token
        self.message = message
        self.message_kind = message_kind

    def __str__(self):
        lines = [f"Nested JSON Error {self.message_kind} Error: {self.message}"]
        if self.token is not None:
            lines.append(self.source)
            lines.append(
                " " * self.token.start
                + HIGHLIGHTER * (self.token.end - self.token.start)
            )
        return "\n".join(lines)


def parse(source: str) -> Iterator[Path]:
    """
    start: root_path path*
    root_path: (literal | index_path | append_path)
    literal: TEXT | NUMBER

    path:
        key_path
        | index_path
        | append_path
    key_path: LEFT_BRACKET TEXT RIGHT_BRACKET
    index_path: LEFT_BRACKET NUMBER RIGHT_BRACKET
    append_path: LEFT_BRACKET RIGHT_BRACKET

    """

    tokens = list(tokenize(source))
    cursor = 0

    def can_advance():
        return cursor < len(tokens)

    # noinspection PyShadowingNames
    def expect(*kinds):
        nonlocal cursor
        assert kinds
        if can_advance():
            token = tokens[cursor]
            cursor += 1
            if token.kind in kinds:
                return token
        elif tokens:
            token = tokens[-1]._replace(
                start=tokens[-1].end + 0,
                end=tokens[-1].end + 1,
            )
        else:
            token = None
        if len(kinds) == 1:
            suffix = kinds[0].to_name()
        else:
            suffix = ", ".join(kind.to_name() for kind in kinds[:-1])
            suffix += " or " + kinds[-1].to_name()
        message = f"Expecting {suffix}"
        raise NestedJSONSyntaxError(source, token, message)

    # noinspection PyShadowingNames
    def parse_root():
        tokens = []
        if not can_advance():
            return Path(kind=PathAction.KEY, accessor=EMPTY_STRING, is_root=True)
        # (literal | index_path | append_path)?
        token = expect(*LITERAL_TOKENS, TokenKind.LEFT_BRACKET)
        tokens.append(token)
        if token.kind in LITERAL_TOKENS:
            action = PathAction.KEY
            value = str(token.value)
        elif token.kind is TokenKind.LEFT_BRACKET:
            token = expect(TokenKind.NUMBER, TokenKind.RIGHT_BRACKET)
            tokens.append(token)
            if token.kind is TokenKind.NUMBER:
                action = PathAction.INDEX
                value = token.value
                tokens.append(expect(TokenKind.RIGHT_BRACKET))
            elif token.kind is TokenKind.RIGHT_BRACKET:
                action = PathAction.APPEND
                value = None
            else:
                assert_cant_happen()
        else:
            assert_cant_happen()
        # noinspection PyUnboundLocalVariable
        return Path(kind=action, accessor=value, tokens=tokens, is_root=True)

    yield parse_root()

    # path*
    while can_advance():
        path_tokens = [expect(TokenKind.LEFT_BRACKET)]
        token = expect(TokenKind.TEXT, TokenKind.NUMBER, TokenKind.RIGHT_BRACKET)
        path_tokens.append(token)
        if token.kind is TokenKind.RIGHT_BRACKET:
            path = Path(PathAction.APPEND, tokens=path_tokens)
        elif token.kind is TokenKind.TEXT:
            path = Path(PathAction.KEY, token.value, tokens=path_tokens)
            path_tokens.append(expect(TokenKind.RIGHT_BRACKET))
        elif token.kind is TokenKind.NUMBER:
            path = Path(PathAction.INDEX, token.value, tokens=path_tokens)
            path_tokens.append(expect(TokenKind.RIGHT_BRACKET))
        else:
            assert_cant_happen()
        # noinspection PyUnboundLocalVariable
        yield path


def tokenize(source: str) -> Iterator[Token]:
    cursor = 0
    backslashes = 0
    buffer: list[str] = []

    def send_buffer() -> Iterator[Token]:
        nonlocal backslashes
        if not buffer:
            return None

        value = "".join(buffer)
        kind = TokenKind.TEXT
        if not backslashes:
            for variation, kind in [
                (int, TokenKind.NUMBER),
                (check_escaped_int, TokenKind.TEXT),
            ]:
                try:
                    value = variation(value)  # type: ignore[operator]
                except ValueError:
                    continue
                else:
                    break
        yield Token(
            kind=kind,
            value=value,
            start=cursor - (len(buffer) + backslashes),
            end=cursor,
        )
        buffer.clear()
        backslashes = 0

    def can_advance() -> bool:
        return cursor < len(source)

    while can_advance():
        index = source[cursor]
        if index in OPERATORS:
            yield from send_buffer()
            yield Token(OPERATORS[index], index, cursor, cursor + 1)
        elif index == BACKSLASH and can_advance():
            if source[cursor + 1] in SPECIAL_CHARS:
                backslashes += 1
            else:
                buffer.append(index)
            buffer.append(source[cursor + 1])
            cursor += 1
        else:
            buffer.append(index)
        cursor += 1

    yield from send_buffer()


def check_escaped_int(value: str) -> str:
    if not value.startswith(BACKSLASH):
        raise ValueError("Not an escaped int")
    try:
        int(value[1:])
    except ValueError as exc:
        raise ValueError("Not an escaped int") from exc
    else:
        return value[1:]


def assert_cant_happen():
    raise ValueError("Unexpected value")


JSONType = type[Union[dict, list, int, float, str]]
JSON_TYPE_MAPPING = {
    dict: "object",
    list: "array",
    int: "number",
    float: "number",
    str: "string",
}


def interpret_nested_json(pairs: Iterable[tuple[str, str]]) -> dict:
    context = None
    for key, value in pairs:
        context = interpret(context, key, value)
    return wrap_with_dict(context)


def interpret(context: Any, key: str, value: Any) -> Any:
    cursor = context
    paths = list(parse(key))
    paths.append(Path(PathAction.SET, value))

    # noinspection PyShadowingNames
    def type_check(index: int, path: Path, expected_type: JSONType):
        if not isinstance(cursor, expected_type):
            if path.tokens:
                pseudo_token = Token(
                    kind=TokenKind.PSEUDO,
                    value="",
                    start=path.tokens[0].start,
                    end=path.tokens[-1].end,
                )
            else:
                pseudo_token = None
            cursor_type = JSON_TYPE_MAPPING.get(type(cursor), type(cursor).__name__)
            required_type = JSON_TYPE_MAPPING[expected_type]
            message = f"Cannot perform {path.kind.to_string()!r} based access on "
            message += repr("".join(path.reconstruct() for path in paths[:index]))
            message += f" which has a type of {cursor_type!r} but this operation"
            message += f" requires a type of {required_type!r}."
            raise NestedJSONSyntaxError(
                source=key,
                token=pseudo_token,
                message=message,
                message_kind="Type",
            )

    def object_for(kind: PathAction) -> Any:
        if kind is PathAction.KEY:
            return {}
        elif kind in {PathAction.INDEX, PathAction.APPEND}:
            return []
        else:
            assert_cant_happen()

    for index, (path, next_path) in enumerate(zip(paths, paths[1:])):
        # If there is no context yet, set it.
        if cursor is None:
            context = cursor = object_for(path.kind)
        if path.kind is PathAction.KEY:
            type_check(index, path, dict)
            if next_path.kind is PathAction.SET:
                cursor[path.accessor] = next_path.accessor
                break
            cursor = cursor.setdefault(path.accessor, object_for(next_path.kind))
        elif path.kind is PathAction.INDEX:
            type_check(index, path, list)
            assert isinstance(path.accessor, int)
            if path.accessor < 0:
                raise NestedJSONSyntaxError(
                    source=key,
                    token=path.tokens[1],
                    message="Negative indexes are not supported.",
                    message_kind="Value",
                )
            cursor.extend([None] * (path.accessor - len(cursor) + 1))
            if next_path.kind is PathAction.SET:
                cursor[path.accessor] = next_path.accessor
                break
            if cursor[path.accessor] is None:
                cursor[path.accessor] = object_for(next_path.kind)
            cursor = cursor[path.accessor]
        elif path.kind is PathAction.APPEND:
            type_check(index, path, list)
            if next_path.kind is PathAction.SET:
                cursor.append(next_path.accessor)
                break
            cursor.append(object_for(next_path.kind))
            cursor = cursor[-1]
        else:
            assert_cant_happen()

    return context


def wrap_with_dict(context):
    if context is None:
        return {}
    elif isinstance(context, list):
        return {
            EMPTY_STRING: NestedJSONArray(context),
        }
    else:
        assert isinstance(context, dict)
        return context


def unwrap_top_level_list_if_needed(data: dict):
    """
    Propagate the top-level list, if that's what we got.

    """
    if len(data) == 1:
        key, value = list(data.items())[0]
        if isinstance(value, NestedJSONArray):
            assert key == EMPTY_STRING
            return value
    return data



from __future__ import annotations

import json
import time
from contextlib import contextmanager
from dataclasses import dataclass, field
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Iterator

import httpx

from fal import flags
from fal.sdk import Credentials, get_default_credentials

if TYPE_CHECKING:
    from websockets.sync.connection import Connection

_STREAM_URL_FORMAT = f"https://{flags.FAL_RUN_HOST}/{{app_id}}"
_QUEUE_URL_FORMAT = f"https://{flags.FAL_QUEUE_RUN_HOST}/{{app_id}}"
_REALTIME_URL_FORMAT = f"wss://{flags.FAL_RUN_HOST}/{{app_id}}"
_WS_URL_FORMAT = f"wss://ws.{flags.FAL_RUN_HOST}/{{app_id}}"


def _backwards_compatible_app_id(app_id: str) -> str:
    if "/" not in app_id:
        # Convert the app_id to the format used in the URL.
        return app_id.replace("-", "/", 1)

    return app_id


@dataclass
class _Status: ...


@dataclass
class Queued(_Status):
    """Indicates the request is still in the queue, and provides the position
    in the queue for ETA calculation."""

    position: int


@dataclass
class InProgress(_Status):
    """Indicates the request is now being actively processed, and provides runtime
    logs for the inference task."""

    logs: list[dict[str, Any]] | None = field()


@dataclass
class Completed(_Status):
    """Indicates the request has been completed successfully and the result is
    ready to be retrieved."""

    logs: list[dict[str, Any]] | None = field()


@lru_cache(maxsize=1)
def _get_http_client() -> httpx.Client:
    return httpx.Client(headers={"User-Agent": "Fal/Python"})


@dataclass
class RequestHandle:
    """A handle to an async inference request."""

    app_id: str
    request_id: str

    _client: httpx.Client = field(default_factory=_get_http_client)

    # Use the credentials that were used to submit the request by default.
    _creds: Credentials = field(default_factory=get_default_credentials, repr=False)

    def __post_init__(self):
        app_id = _backwards_compatible_app_id(self.app_id)
        # drop any extra path components
        parts = app_id.split("/")[:3]
        if parts[0] != "workflows":
            # if the app_id is not a workflow, only keep the first two parts
            parts = parts[:2]

        self.app_id = "/".join(parts)

    def status(self, *, logs: bool = False) -> _Status:
        """Check the status of an async inference request."""

        url = (
            _QUEUE_URL_FORMAT.format(app_id=self.app_id)
            + f"/requests/{self.request_id}/status/"
        )
        response = self._client.get(
            url,
            headers=self._creds.to_headers(),
            params={"logs": int(logs)},
        )
        response.raise_for_status()

        data = response.json()

        if response.status_code == 200:
            return Completed(logs=data["logs"])

        if data["status"] == "IN_QUEUE":
            return Queued(position=data["queue_position"])
        elif data["status"] == "IN_PROGRESS":
            return InProgress(logs=data["logs"])
        else:
            raise ValueError(f"Unknown status: {data['status']}")

    def cancel(self) -> None:
        """Cancel an async inference request."""
        url = (
            _QUEUE_URL_FORMAT.format(app_id=self.app_id)
            + f"/requests/{self.request_id}/cancel"
        )
        response = self._client.put(url, headers=self._creds.to_headers())
        response.raise_for_status()

    def iter_events(
        self,
        *,
        logs: bool = False,
        __poll_delay: float = 0.2,
    ) -> Iterator[_Status]:
        """Yield all events regarding the given task till its completed."""

        while True:
            status = self.status(logs=logs)

            if isinstance(status, Completed):
                return

            yield status
            time.sleep(__poll_delay)

    def fetch_raw_response(self) -> httpx.Response:
        url = (
            _QUEUE_URL_FORMAT.format(app_id=self.app_id)
            + f"/requests/{self.request_id}/"
        )
        response = self._client.get(url, headers=self._creds.to_headers())
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if "application/json" not in response.headers["Content-Type"]:
                raise
            raise httpx.HTTPStatusError(
                f"{response.status_code}: {response.text}",
                request=e.request,
                response=e.response,
            ) from e
        return response

    def fetch_result(self) -> dict[str, Any]:
        """Retrieve the result of an async inference request, raises an exception
        if the request is not completed yet."""
        response = self.fetch_raw_response()
        return response.json()

    def get(self) -> dict[str, Any]:
        """Retrieve the result of an async inference request, polling the status
        of the request until it is completed."""

        for _event in self.iter_events(logs=False):
            continue

        return self.fetch_result()


def stream(
    app_id: str, arguments: dict[str, Any], *, path: str = ""
) -> Iterator[str | bytes]:
    """Stream an inference task on a Fal app."""

    app_id = _backwards_compatible_app_id(app_id)
    url = _STREAM_URL_FORMAT.format(app_id=app_id)
    if path:
        _path = path[len("/") :] if path.startswith("/") else path
        url += "/" + _path

    creds = get_default_credentials()
    client = _get_http_client()

    response = client.post(
        url,
        json=arguments,
        headers=creds.to_headers(),
    )
    response.raise_for_status()

    if response.headers["Content-Type"].startswith("text/event-stream"):
        for line in response.iter_lines():
            if line:
                yield line
    else:
        yield from response.iter_bytes()


def run(app_id: str, arguments: dict[str, Any], *, path: str = "") -> dict[str, Any]:
    """Run an inference task on a Fal app and return the result."""

    handle = submit(app_id, arguments, path=path)
    return handle.get()


def submit(app_id: str, arguments: dict[str, Any], *, path: str = "") -> RequestHandle:
    """Submit an async inference task to the app. Returns a request handle
    which can be used to check the status of the request and retrieve the
    result."""

    app_id = _backwards_compatible_app_id(app_id)
    url = _QUEUE_URL_FORMAT.format(app_id=app_id)
    if path:
        _path = path[len("/") :] if path.startswith("/") else path
        url += "/" + _path

    creds = get_default_credentials()
    client = _get_http_client()

    response = client.post(
        url,
        json=arguments,
        headers=creds.to_headers(),
    )
    response.raise_for_status()

    data = response.json()
    return RequestHandle(
        app_id=app_id,
        request_id=data["request_id"],
        _creds=creds,
        _client=client,
    )


@dataclass
class _RealtimeConnection:
    """A realtime connection to a Fal app."""

    _ws: Any

    def run(self, arguments: dict[str, Any]) -> dict[str, Any]:
        """Run an inference task on the app and return the result."""
        self.send(arguments)
        return self.recv()

    def send(self, arguments: dict[str, Any]) -> None:
        import msgpack

        """Send an inference task to the app."""
        payload = msgpack.packb(arguments)
        self._ws.send(payload)

    def recv(self) -> dict[str, Any]:
        import msgpack

        """Receive the result of an inference task."""
        while True:
            response = self._ws.recv()
            if isinstance(response, str):
                print(response)
                json_payload = json.loads(response)
                if json_payload.get("type") == "x-fal-error":
                    raise ValueError(json_payload["reason"])
                continue
            return msgpack.unpackb(response)


@contextmanager
def _connect(app_id: str, *, path: str = "/realtime") -> Iterator[_RealtimeConnection]:
    """Connect to a realtime endpoint. This is an internal and experimental API, use it
    at your own risk."""

    from websockets.sync import client

    app_id = _backwards_compatible_app_id(app_id)
    url = _REALTIME_URL_FORMAT.format(app_id=app_id)
    if path:
        _path = path[len("/") :] if path.startswith("/") else path
        url += "/" + _path

    creds = get_default_credentials()

    with client.connect(
        url, additional_headers=creds.to_headers(), open_timeout=90
    ) as ws:
        yield _RealtimeConnection(ws)


class _MetaMessageFound(Exception): ...


@dataclass
class _WSConnection:
    """A WS connection to an HTTP Fal app."""

    _ws: Connection
    _buffer: str | bytes | None = None

    def run(self, arguments: dict[str, Any]) -> bytes:
        """Run an inference task on the app and return the result."""
        self.send(arguments)
        return self.recv()

    def send(self, arguments: dict[str, Any]) -> None:
        import json

        payload = json.dumps(arguments)
        self._ws.send(payload)

    def _peek(self) -> bytes | str:
        if self._buffer is None:
            self._buffer = self._ws.recv()

        return self._buffer

    def _consume(self) -> None:
        if self._buffer is None:
            raise ValueError("No data to consume")

        self._buffer = None

    @contextmanager
    def _recv(self) -> Iterator[str | bytes]:
        res = self._peek()

        yield res

        # Only consume if it went through the context manager without raising
        self._consume()

    def _is_meta(self, res: str | bytes) -> bool:
        if not isinstance(res, str):
            return False

        try:
            json_payload: Any = json.loads(res)
        except json.JSONDecodeError:
            return False

        if not isinstance(json_payload, dict):
            return False

        return "type" in json_payload and "request_id" in json_payload

    def _recv_meta(self, type: str) -> dict[str, Any]:
        with self._recv() as res:
            if not self._is_meta(res):
                raise ValueError(f"Expected a {type} message")

            json_payload: dict = json.loads(res)
            if json_payload.get("type") != type:
                raise ValueError(f"Expected a {type} message")

            return json_payload

    def _recv_response(self) -> Iterator[str | bytes]:
        while True:
            try:
                with self._recv() as res:
                    if self._is_meta(res):
                        # Raise so we dont consume the message
                        raise _MetaMessageFound()

                    yield res
            except _MetaMessageFound:
                break

    def recv(self) -> bytes:
        start = self._recv_meta("start")
        request_id = start["request_id"]

        response = b""
        for part in self._recv_response():
            if isinstance(part, str):
                response += part.encode()
            else:
                response += part

        end = self._recv_meta("end")
        if end["request_id"] != request_id:
            raise ValueError("Mismatched request_id in end message")

        return response

    def stream(self) -> Iterator[str | bytes]:
        start = self._recv_meta("start")
        request_id = start["request_id"]

        yield from self._recv_response()

        # Make sure we consume the end message
        end = self._recv_meta("end")
        if end["request_id"] != request_id:
            raise ValueError("Mismatched request_id in end message")


@contextmanager
def ws(app_id: str, *, path: str = "") -> Iterator[_WSConnection]:
    """Connect to a HTTP endpoint but with websocket protocol. This is an internal and
    experimental API, use it at your own risk."""

    from websockets.sync import client

    app_id = _backwards_compatible_app_id(app_id)
    url = _WS_URL_FORMAT.format(app_id=app_id)
    if path:
        _path = path[len("/") :] if path.startswith("/") else path
        url += "/" + _path

    creds = get_default_credentials()

    with client.connect(
        url, additional_headers=creds.to_headers(), open_timeout=90
    ) as ws:
        yield _WSConnection(ws)



import pytest
from pydantic import BaseModel

import fal
from fal.app import AppClient
from fal.distributed.worker import DistributedRunner, DistributedWorker


class EchoRequest(BaseModel):
    message: str


class EchoResponse(BaseModel):
    echo: str


class EchoWorker(DistributedWorker):
    def __call__(self, message: str = "", **kwargs):
        # Accept additional kwargs like 'streaming' that the runner may pass
        return {"echo": message}


class EchoApp(fal.App):
    # Specify requirements and machine type for remote deployment
    machine_type = "H100"
    num_gpus = 1  # Request a GPU machine for distributed processing
    keep_alive = 300
    requirements = [
        "pyzmq",  # Required for distributed communication
        "torch",  # Required for distributed processing
    ]

    async def setup(self):
        self.runner = DistributedRunner(EchoWorker, world_size=1)
        await self.runner.start()

    @fal.endpoint("/")
    async def echo(self, request: EchoRequest) -> EchoResponse:
        result = await self.runner.invoke({"message": request.message})
        # Handle potential error in result
        if "error" in result:
            raise RuntimeError(f"Worker error: {result['error']}")
        return EchoResponse(echo=result.get("echo", ""))


@pytest.fixture(scope="module")
def echo_app_client():
    with AppClient.connect(EchoApp) as client:
        yield client


@pytest.mark.skip(reason="Requires GPU machine")
def test_echo_endpoint(echo_app_client):
    """Test that the distributed echo app works correctly on remote GPU.

    This test verifies:
    1. The DistributedRunner can start on a remote GPU
    2. Workers can process requests
    3. Results are returned correctly
    """
    response = echo_app_client.echo(EchoRequest(message="hello"))
    assert response.echo == "hello"

    # Test with different message
    response2 = echo_app_client.echo(EchoRequest(message="world"))
    assert response2.echo == "world"



from __future__ import annotations

import json
from contextlib import contextmanager
from datetime import datetime, timezone
from unittest.mock import MagicMock, patch

import pytest

from fal.toolkit.file.providers import fal as providers
from fal.toolkit.file.types import FileData


@contextmanager
def _fake_retry_request(_request, **_kwargs):
    yield MagicMock()


def _assert_lifecycle_headers(
    headers: dict[str, str], preference: dict[str, str]
) -> None:
    expected = json.dumps(preference)
    assert headers["X-Fal-Object-Lifecycle"] == expected
    assert headers["X-Fal-Object-Lifecycle-Preference"] == expected


def _assert_multipart_preference_call(mock, preference: dict[str, str]) -> None:
    _, kwargs = mock.call_args
    assert kwargs["object_lifecycle_preference"] == preference


@pytest.mark.parametrize(
    ("multipart_cls", "token_manager", "token", "json_payload"),
    [
        (
            providers.MultipartUploadGCS,
            None,
            None,
            {"file_url": "https://files", "upload_url": "https://upload"},
        ),
        (
            providers.MultipartUpload,
            providers.fal_v2_token_manager,
            providers.FalV2Token(
                token="token",
                token_type="Bearer",
                base_upload_url="https://upload",
                expires_at=datetime.now(timezone.utc),
            ),
            {"file_url": "https://files", "upload_url": "https://upload"},
        ),
        (
            providers.MultipartUploadV3,
            None,
            None,
            {"file_url": "https://files", "upload_url": "https://upload"},
        ),
        (
            providers.InternalMultipartUploadV3,
            providers.fal_v3_token_manager,
            providers.FalV3Token(
                token="token",
                token_type="Bearer",
                base_upload_url="https://upload",
                expires_at=datetime.now(timezone.utc),
            ),
            {"access_url": "https://files", "uploadId": "upload-id"},
        ),
    ],
)
def test_multipart_create_includes_lifecycle_headers(
    multipart_cls, token_manager, token, json_payload
):
    captured: dict[str, dict[str, str]] = {}

    def _fake_request(_url, headers=None, method=None, data=None):
        captured["headers"] = headers or {}
        return MagicMock()

    patches = [
        patch.object(providers, "Request", side_effect=_fake_request),
        patch.object(providers, "_maybe_retry_request", _fake_retry_request),
        patch.object(providers.json, "load", return_value=json_payload),
    ]
    if token_manager:
        patches.append(patch.object(token_manager, "get_token", return_value=token))
    else:
        patches.append(
            patch.object(
                providers, "key_credentials", return_value=("key_id", "key_secret")
            )
        )

    with patches[0], patches[1], patches[2], patches[3]:
        multipart = multipart_cls("file.txt")
        multipart.create(object_lifecycle_preference={"ttl": "1d"})

    _assert_lifecycle_headers(captured["headers"], {"ttl": "1d"})


@pytest.mark.parametrize(
    ("repo_cls", "multipart_cls", "method_name"),
    [
        (providers.FalFileRepository, providers.MultipartUploadGCS, "save"),
        (providers.FalFileRepository, providers.MultipartUploadGCS, "save_file"),
        (providers.FalFileRepositoryV2, providers.MultipartUpload, "save"),
        (providers.FalFileRepositoryV2, providers.MultipartUpload, "save_file"),
        (providers.FalFileRepositoryV3, providers.MultipartUploadV3, "save"),
        (providers.FalFileRepositoryV3, providers.MultipartUploadV3, "save_file"),
        (
            providers.InternalFalFileRepositoryV3,
            providers.InternalMultipartUploadV3,
            "save",
        ),
        (
            providers.InternalFalFileRepositoryV3,
            providers.InternalMultipartUploadV3,
            "save_file",
        ),
    ],
)
def test_repository_multipart_passes_lifecycle_preference(
    repo_cls, multipart_cls, method_name
):
    preference = {"ttl": "1d"}
    repo = repo_cls()

    with patch.object(multipart_cls, method_name, return_value="url") as mock:
        if method_name == "save":
            file = FileData(b"data", content_type="text/plain", file_name="file.txt")
            repo.save(file, multipart=True, object_lifecycle_preference=preference)
        else:
            repo.save_file(
                "path/to/file.txt",
                content_type="text/plain",
                multipart=True,
                object_lifecycle_preference=preference,
            )

    _assert_multipart_preference_call(mock, preference)



from __future__ import annotations

from base64 import b64encode
from io import BytesIO
from typing import Literal, overload

import pytest
from PIL import Image as PILImage
from pydantic import BaseModel, Field
from pydantic import __version__ as pydantic_version

from fal.toolkit import Image


@overload
def get_image(as_bytes: Literal[False] = False) -> PILImage.Image: ...


@overload
def get_image(as_bytes: Literal[True]) -> bytes: ...


def get_image(as_bytes: bool = False):
    from PIL import Image

    pil_image = Image.new("RGB", (1, 1), (255, 255, 255))
    if not as_bytes:
        return pil_image

    return pil_image_to_bytes(pil_image)


def pil_image_to_bytes(image: PILImage.Image) -> bytes:
    image_bytes = BytesIO()
    image.save(image_bytes, format="PNG")
    return image_bytes.getvalue()


def fal_image_downloaded(image: Image):
    return image.file_size is not None


def fal_image_url_matches(image: Image, url: str):
    return image.url == url


def fal_image_content_matches(image: Image, content: bytes):
    image1 = PILImage.open(BytesIO(image.as_bytes()))
    image2 = PILImage.open(BytesIO(content))
    return images_are_equal(image1, image2)


def image_to_data_uri(image: PILImage.Image) -> str:
    image_bytes = pil_image_to_bytes(image)
    b64_encoded = b64encode(image_bytes).decode("utf-8")
    return f"data:image/png;base64,{b64_encoded}"


def images_are_equal(img1: PILImage.Image, img2: PILImage.Image) -> bool:
    pixels1 = list(img1.getdata())
    pixels2 = list(img2.getdata())
    return pixels1 == pixels2


def assert_fal_images_equals(fal_image_1: Image, fal_image_2: Image):
    assert fal_image_content_matches(fal_image_1, fal_image_2.as_bytes())

    assert (
        fal_image_1.file_size == fal_image_2.file_size
    ), "Image file size should match"
    assert (
        fal_image_1.content_type == fal_image_2.content_type
    ), "Content type should match"
    assert fal_image_1.url == fal_image_2.url, "URL should match"
    assert fal_image_1.width == fal_image_2.width, "Width should match"
    assert fal_image_1.height == fal_image_2.height, "Height should match"


def test_fal_image_from_pil(isolated_client):
    @isolated_client(requirements=["pillow", f"pydantic=={pydantic_version}", "tomli"])
    def fal_image_from_bytes_remote():
        pil_image = get_image()
        return Image.from_pil(pil_image, repository="in_memory")

    fal_image = fal_image_from_bytes_remote()
    assert fal_image_content_matches(fal_image, get_image(as_bytes=True))


def test_fal_image_from_bytes(isolated_client):
    @isolated_client(requirements=["pillow", f"pydantic=={pydantic_version}", "tomli"])
    def fal_image_from_bytes_remote():
        image_bytes = get_image(as_bytes=True)
        return Image.from_bytes(image_bytes, repository="in_memory", format="png")

    fal_image = fal_image_from_bytes_remote()
    assert fal_image_content_matches(fal_image, get_image(as_bytes=True))


@pytest.mark.parametrize(
    "image_url",
    [
        "https://commons.wikimedia.org/static/images/project-logos/commonswiki-1.5x.png",
        image_to_data_uri(get_image()),
    ],
)
def test_fal_image_input(isolated_client, image_url):
    class TestInput(BaseModel):
        image: Image = Field()

    @isolated_client(requirements=["pillow", f"pydantic=={pydantic_version}", "tomli"])
    def init_image_on_fal(input: TestInput) -> Image:
        return TestInput(image=input.image).image

    test_input = TestInput(image=image_url)
    image = init_image_on_fal(test_input)

    # Image is not downloaded until it is needed
    assert not fal_image_downloaded(image)
    assert fal_image_url_matches(image, image_url)

    # Expect value error if we try to access the file content for input file
    with pytest.raises(ValueError):
        image.as_bytes()


def test_fal_image_input_to_pil(isolated_client):
    class TestInput(BaseModel):
        image: Image = Field()

    @isolated_client(requirements=["pillow", f"pydantic=={pydantic_version}", "tomli"])
    def init_image_on_fal(input: TestInput) -> bytes:
        input_image = TestInput(image=input.image).image
        pil_image = input_image.to_pil()
        return pil_image_to_bytes(pil_image)

    test_input = TestInput(image=Image.from_pil(get_image()))
    image_bytes = init_image_on_fal(test_input)

    assert image_bytes == get_image(as_bytes=True)



from __future__ import annotations

from ._base import FalServerlessException


class UnauthenticatedException(FalServerlessException):
    def __init__(self) -> None:
        super().__init__(
            "You must be authenticated. "
            "Use [bold]fal auth login[/] or [bold]fal profile key[/] to set your fal key."  # noqa: E501
        )



import re
import tempfile
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

import fal.file_sync as file_sync_mod
from fal.file_sync import FileSync


@pytest.fixture
def temp_dir():
    """Create a temporary directory for tests"""
    with tempfile.TemporaryDirectory() as temp_dir:
        yield temp_dir


@pytest.fixture
def file_sync(temp_dir, monkeypatch):
    """Create FileSync instance with temporary directory"""
    # Ensure credential and host resolution do not require real auth or HTTPS
    monkeypatch.setenv("FAL_KEY", "dummy:dummy")
    monkeypatch.setenv("FAL_HOST", "api.localhost")
    monkeypatch.setenv("ISOLATE_TEST_MODE", "1")
    return FileSync(local_file_path=str(Path(temp_dir) / "app.py"))


def test_file_sync_init(temp_dir):
    """Test FileSync initialization sets correct paths"""
    local_path = str(Path(temp_dir) / "test_app.py")
    fs = FileSync(local_path)
    assert fs.local_file_path == local_path


def test_request_success(file_sync):
    """Test successful HTTP request returns response"""
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = {"success": True}

    with patch.object(file_sync, "_client") as mock_client:
        mock_client.request.return_value = mock_response
        response = file_sync._request("GET", "/test")
        assert response == mock_response


def test_request_failure_with_json_detail(file_sync):
    """Test HTTP request failure raises exception with JSON error detail"""
    from fal.exceptions import FalServerlessException

    mock_response = MagicMock()
    mock_response.status_code = 400
    mock_response.json.return_value = {"detail": "Bad request"}

    with patch.object(file_sync, "_client") as mock_client:
        mock_client.request.return_value = mock_response
        with pytest.raises(FalServerlessException, match="Bad request"):
            file_sync._request("GET", "/test")


def test_request_404_raises_not_found(file_sync):
    """Test HTTP 404 raises specific not found exception"""
    from fal.exceptions import FalServerlessException

    mock_response = MagicMock()
    mock_response.status_code = 404

    with patch.object(file_sync, "_client") as mock_client:
        mock_client.request.return_value = mock_response
        with pytest.raises(FalServerlessException, match="Not Found"):
            file_sync._request("GET", "/test")


def test_compute_hash_basic_functionality(temp_dir):
    """Test hash computation produces valid SHA256"""
    test_file = Path(temp_dir) / "test.txt"
    test_file.write_text("Hello, world!")

    stat = test_file.stat()
    hash_result = file_sync_mod.compute_hash(test_file, stat.st_mode)

    assert len(hash_result) == 64
    assert all(c in "0123456789abcdef" for c in hash_result)


def test_compute_hash_different_content_different_hash(temp_dir):
    """Test different file content produces different hashes"""
    file1 = Path(temp_dir) / "file1.txt"
    file2 = Path(temp_dir) / "file2.txt"
    file1.write_text("content1")
    file2.write_text("content2")

    hash1 = file_sync_mod.compute_hash(file1, file1.stat().st_mode)
    hash2 = file_sync_mod.compute_hash(file2, file2.stat().st_mode)

    assert hash1 != hash2


def test_get_file_metadata_structure(temp_dir):
    """Test file metadata returns correct structure and values"""
    test_file = Path(temp_dir) / "test.txt"
    test_content = "test content"
    test_file.write_text(test_content)

    metadata = file_sync_mod.FileMetadata.from_path(
        test_file, relative="test.txt", absolute=str(test_file)
    )

    assert metadata.size == len(test_content)
    assert len(metadata.hash) == 64


def test_get_file_metadata_file_too_large_error(temp_dir):
    """Test that files exceeding size limit raise FileTooLargeError"""
    from fal.exceptions import FileTooLargeError

    test_file = Path(temp_dir) / "test.txt"
    test_file.write_text("content")

    # Mock the stat to return a size larger than FILE_SIZE_LIMIT
    with patch("pathlib.Path.stat") as mock_stat:
        mock_stat.return_value.st_size = 1024 * 1024 * 1024 + 1  # 1GB + 1 byte
        mock_stat.return_value.st_mtime = 123456789
        mock_stat.return_value.st_mode = 33188

        with pytest.raises(FileTooLargeError):
            file_sync_mod.FileMetadata.from_path(
                test_file, relative="test.txt", absolute=str(test_file)
            )


def test_normalize_path_relative_to_base(temp_dir):
    """Test path normalization resolves relative paths correctly"""
    base_file = str(Path(temp_dir) / "app.py")

    abs_path, rel_path = file_sync_mod.normalize_path("config.txt", base_file)

    expected_abs = (Path(temp_dir) / "config.txt").resolve().as_posix()
    assert abs_path == expected_abs
    assert rel_path == "config.txt"


def test_collect_files_single_file(temp_dir):
    """Test collecting a single file"""
    app_file = Path(temp_dir) / "app.py"
    app_file.write_text("# app file")

    single_file = Path(temp_dir) / "readme.txt"
    single_file.write_text("readme content")

    fs = FileSync(str(app_file))
    files = fs.collect_files(["readme.txt"])

    assert len(files) == 1
    assert files[0].relative_path == "readme.txt"


def test_collect_files_directory(temp_dir):
    """Test collecting files from a directory"""
    app_file = Path(temp_dir) / "app.py"
    app_file.write_text("# app file")

    config_dir = Path(temp_dir) / "config"
    config_dir.mkdir()
    (config_dir / "config.json").write_text('{"key": "value"}')
    (config_dir / "settings.yml").write_text("key: value")

    fs = FileSync(str(app_file))
    files = fs.collect_files(["config/"])

    assert len(files) == 2
    relative_paths = [f.relative_path for f in files]
    assert "config/config.json" in relative_paths
    assert "config/settings.yml" in relative_paths


def test_collect_files_handles_nonexistent_gracefully(temp_dir):
    """Test that nonexistent files/directories are skipped without error"""
    app_file = Path(temp_dir) / "app.py"
    app_file.write_text("# app file")

    fs = FileSync(str(app_file))
    files = fs.collect_files(["nonexistent.txt", "missing_dir/"])

    assert len(files) == 0


def test_should_ignore_file_basic_patterns(file_sync):
    """Test basic regex patterns for ignoring files"""
    ignore_patterns_str = [r"\.pyc$", r"__pycache__/", r"\.git/"]
    ignore_patterns = [re.compile(pattern) for pattern in ignore_patterns_str]

    # Should ignore
    assert file_sync._matches_patterns("test.pyc", ignore_patterns)
    assert file_sync._matches_patterns("__pycache__/file.py", ignore_patterns)
    assert file_sync._matches_patterns(".git/config", ignore_patterns)

    # Should not ignore
    assert not file_sync._matches_patterns("test.py", ignore_patterns)
    assert not file_sync._matches_patterns("src/main.py", ignore_patterns)


def test_should_ignore_file_empty_patterns(file_sync):
    """Test behavior with empty ignore patterns"""
    result = file_sync._matches_patterns("test.py", [])
    assert not result


def test_check_hashes_on_server_success(file_sync):
    """Test successful hash check on server"""
    test_hashes = ["hash1", "hash2", "hash3"]
    missing_hashes = ["hash2"]

    mock_response = MagicMock()
    mock_response.json.return_value = missing_hashes

    with patch.object(file_sync, "_request", return_value=mock_response):
        result = file_sync.check_hashes_on_server(test_hashes)
        assert result == missing_hashes


def test_check_hashes_on_server_error_returns_all_missing(file_sync):
    """Test hash check handles server errors gracefully"""
    test_hashes = ["hash1", "hash2"]

    with patch.object(file_sync, "_request", side_effect=Exception("Network error")):
        result = file_sync.check_hashes_on_server(test_hashes)
        assert result == test_hashes


def test_upload_file_multipart_basic_functionality(file_sync, temp_dir):
    """Test multipart file upload calls correct methods with parameters"""
    test_file = Path(temp_dir) / "test.txt"
    test_file.write_text("test content")

    metadata = file_sync_mod.FileMetadata.from_path(
        test_file, relative="test.txt", absolute=str(test_file)
    )

    with patch("fal.file_sync.AppFileMultipartUpload") as MockMultipart:
        mock_instance = MagicMock()
        mock_instance.upload_file.return_value = "test_etag"
        MockMultipart.return_value = mock_instance

        result = file_sync.upload_file_multipart(str(test_file), metadata)

        assert result == "test_etag"
        MockMultipart.assert_called_once()
        mock_instance.upload_file.assert_called_once_with(str(test_file))


def test_multipart_upload_returns_md5_etag(temp_dir):
    """Test that multipart upload returns MD5 etag from server"""
    from fal.upload import AppFileMultipartUpload

    test_file = Path(temp_dir) / "test.txt"
    test_file.write_text("test content")

    metadata = file_sync_mod.FileMetadata.from_path(
        test_file, relative="test.txt", absolute=str(test_file)
    )

    mock_client = MagicMock()
    mock_response = MagicMock()
    mock_response.status_code = 200

    # Simulate server responses
    initiate_response = MagicMock()
    initiate_response.status_code = 200
    initiate_response.json.return_value = {"upload_id": "test-upload-id"}

    part_response = MagicMock()
    part_response.status_code = 200
    part_response.json.return_value = {
        "part_number": 1,
        "etag": "d8e8fca2dc0f896fd7cb4cb0031ba249",  # MD5 hash (32 chars)
    }

    complete_response = MagicMock()
    complete_response.status_code = 200
    complete_response.json.return_value = {
        "etag": "d8e8fca2dc0f896fd7cb4cb0031ba249"  # MD5, not SHA256
    }

    mock_client.request.side_effect = [
        initiate_response,
        part_response,
        complete_response,
    ]

    uploader = AppFileMultipartUpload(
        client=mock_client,
        file_hash=metadata.hash,  # SHA256 (64 chars)
        metadata=metadata.to_dict(),
    )

    etag = uploader.upload_file(str(test_file))

    # Server returns MD5 (32 chars), not SHA256 (64 chars)
    assert len(etag) == 32
    assert etag == "d8e8fca2dc0f896fd7cb4cb0031ba249"
    # No client-side hash verification should happen
    # (would fail if we compared MD5 to SHA256)


def test_multipart_upload_bounded_queue_limits_memory(temp_dir):
    """Test that bounded queue prevents loading entire file into memory"""

    from fal.upload import BaseMultipartUpload

    # Create a file with multiple chunks
    test_file = Path(temp_dir) / "large.txt"
    test_file.write_bytes(b"x" * (10 * 1024 * 1024 + 100))  # ~10MB

    mock_client = MagicMock()

    initiate_response = MagicMock()
    initiate_response.status_code = 200
    initiate_response.json.return_value = {"upload_id": "test-id"}

    part_response = MagicMock()
    part_response.status_code = 200
    part_response.json.return_value = {"part_number": 1, "etag": "abc"}

    complete_response = MagicMock()
    complete_response.status_code = 200
    complete_response.json.return_value = {"etag": "final"}

    mock_client.request.side_effect = [
        initiate_response,
        part_response,
        part_response,
        complete_response,
    ]

    class TestUpload(BaseMultipartUpload):
        @property
        def initiate_url(self):
            return "/init"

        @property
        def part_url(self):
            return "/part"

        @property
        def complete_url(self):
            return "/complete"

    uploader = TestUpload(
        client=mock_client,
        chunk_size=10 * 1024 * 1024,  # 10MB
        max_concurrency=2,
    )

    # Should not raise MemoryError even with large file
    etag = uploader.upload_file(str(test_file))
    assert etag == "final"


def test_sync_files_with_ignore_patterns(file_sync, temp_dir):
    """Test sync_files properly filters files using ignore patterns"""
    # Create test files
    app_file = Path(temp_dir) / "app.py"
    app_file.write_text("# app file")

    pyc_file = Path(temp_dir) / "test.pyc"
    pyc_file.write_text("bytecode")

    fs = FileSync(str(app_file))

    # Mock server responses to avoid actual network calls
    with patch.object(fs, "check_hashes_on_server", return_value=[]):
        all_files, errors = fs.sync_files(
            paths=[str(temp_dir)], files_ignore=[re.compile(r"\.pyc$")]
        )

        # Should only process app.py, not the ignored .pyc file
        relative_paths = [f.relative_path for f in all_files]

        assert "app.py" in relative_paths
        assert "test.pyc" not in relative_paths
        assert len(errors) == 0


def test_sync_files_basic_workflow(file_sync, temp_dir):
    """Test basic sync_files workflow"""
    test_file = Path(temp_dir) / "test.txt"
    test_file.write_text("content")

    fs = FileSync(str(Path(temp_dir) / "app.py"))

    # Mock server to indicate all files need upload
    with patch.object(fs, "check_hashes_on_server") as mock_check, patch.object(
        fs, "upload_file_multipart", return_value="test_etag"
    ):
        # Make server say all hashes are missing (need upload)
        mock_check.side_effect = lambda hashes: hashes

        all_files, errors = fs.sync_files([str(test_file)])

        assert len(all_files) >= 1
        assert len(errors) == 0


def test_sync_files_deduplication(file_sync, temp_dir):
    """Test that duplicate absolute paths are deduplicated"""
    test_file = Path(temp_dir) / "test.txt"
    test_file.write_text("content")

    fs = FileSync(str(Path(temp_dir) / "app.py"))

    with patch.object(fs, "check_hashes_on_server", return_value=[]):
        all_files, errors = fs.sync_files([str(test_file), str(test_file)])

        # Should only process the file once
        total_files = len(all_files)
        assert total_files == 1
        assert len(errors) == 0


def test_close_client(file_sync):
    """Test HTTP client cleanup"""
    mock_client = MagicMock()

    with patch.object(file_sync, "_client", mock_client):
        file_sync.close()
        mock_client.close.assert_called_once()


def test_multipart_upload_empty_file(temp_dir):
    """Test that multipart upload handles empty files correctly"""
    from fal.upload import AppFileMultipartUpload

    # Create empty file
    test_file = Path(temp_dir) / "__init__.py"
    test_file.write_bytes(b"")

    metadata = file_sync_mod.FileMetadata.from_path(
        test_file, relative="__init__.py", absolute=str(test_file)
    )

    mock_client = MagicMock()

    initiate_response = MagicMock()
    initiate_response.status_code = 200
    initiate_response.json.return_value = {"upload_id": "test-upload-id"}

    part_response = MagicMock()
    part_response.status_code = 200
    part_response.json.return_value = {
        "part_number": 1,
        "etag": "d41d8cd98f00b204e9800998ecf8427e",
    }

    complete_response = MagicMock()
    complete_response.status_code = 200
    complete_response.json.return_value = {"etag": "d41d8cd98f00b204e9800998ecf8427e"}

    mock_client.request.side_effect = [
        initiate_response,
        part_response,
        complete_response,
    ]

    uploader = AppFileMultipartUpload(
        client=mock_client,
        file_hash=metadata.hash,
        metadata=metadata.to_dict(),
    )

    etag = uploader.upload_file(str(test_file))

    assert etag == "d41d8cd98f00b204e9800998ecf8427e"
    assert mock_client.request.call_count == 3  # initiate, part, complete



from __future__ import annotations

import os
import pickle
from contextvars import ContextVar

import pytest

import fal
from fal import App, endpoint
from fal.container import ContainerImage


class PickleApp(App):
    pass


def test_app_regions_propagate_to_function_options():
    from fal.app import wrap_app

    class RegionsApp(App):
        regions = ["us-east", "eu-west"]

        @endpoint("/")
        def hello(self) -> str:
            return "Hello, world!"

    fn = wrap_app(RegionsApp)
    assert fn.options.host.get("regions") == ["us-east", "eu-west"]


def test_app_default_app_name_is_generated_from_class_name():
    class MyCustomApp(App):
        pass

    assert MyCustomApp.app_name == "my-custom-app"


# ============================================================================
# Tests for _to_fal_app_name - the function that converts class names to app names
# ============================================================================


class TestToFalAppName:
    """Comprehensive tests for _to_fal_app_name conversion logic."""

    # --- Standard PascalCase conversions ---

    def test_simple_pascal_case(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("MyApp") == "my-app"
        assert _to_fal_app_name("SimpleApp") == "simple-app"
        assert _to_fal_app_name("MyGoodApp") == "my-good-app"

    def test_single_word_pascal_case(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("App") == "app"
        assert _to_fal_app_name("Model") == "model"

    def test_long_pascal_case(self):
        from fal.app import _to_fal_app_name

        assert (
            _to_fal_app_name("MyVeryLongApplicationName")
            == "my-very-long-application-name"
        )

    # --- Acronym handling (known quirky behavior) ---

    def test_all_caps_acronym_splits_each_letter(self):
        """Acronyms like SDXL get split into individual letters - known behavior."""
        from fal.app import _to_fal_app_name

        # Each uppercase letter becomes its own segment
        assert _to_fal_app_name("SDXLModel") == "s-d-x-l-model"
        assert _to_fal_app_name("ONNXRuntime") == "o-n-n-x-runtime"
        assert _to_fal_app_name("GPT4App") == "g-p-t-app"  # 4 is not captured by regex

    def test_mixed_acronym_and_words(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("MySDXLApp") == "my-s-d-x-l-app"

    # --- snake_case fallback ---

    def test_snake_case_conversion(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("mock_model") == "mock-model"
        assert _to_fal_app_name("my_cool_app") == "my-cool-app"
        assert _to_fal_app_name("image_generator") == "image-generator"

    def test_snake_case_with_numbers(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("sdxl_v2") == "sdxl-v2"
        assert _to_fal_app_name("gpt_4_turbo") == "gpt-4-turbo"

    def test_snake_case_with_leading_trailing_underscores(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("_private_app") == "private-app"
        assert _to_fal_app_name("app_") == "app"
        assert _to_fal_app_name("__dunder__") == "dunder"

    def test_snake_case_with_multiple_underscores(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("my__app") == "my-app"

    # --- Lowercase fallback ---

    def test_simple_lowercase(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("myapp") == "myapp"
        assert _to_fal_app_name("simple") == "simple"

    def test_lowercase_with_numbers(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("app123") == "app123"
        assert _to_fal_app_name("v2model") == "v2model"

    # --- Edge cases ---

    def test_single_character(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("A") == "a"
        assert _to_fal_app_name("x") == "x"

    def test_numbers_only_falls_through_to_lowercase(self):
        from fal.app import _to_fal_app_name

        # No uppercase letters, no underscores, so falls through to lowercase
        assert _to_fal_app_name("123") == "123"

    def test_mixed_case_not_pascal(self):
        """Names that start lowercase but have uppercase later."""
        from fal.app import _to_fal_app_name

        # Only captures uppercase-starting segments
        assert _to_fal_app_name("myApp") == "app"  # 'my' is ignored, only 'App' matches

    # --- Error cases ---

    def test_empty_string_raises_value_error(self):
        from fal.app import _to_fal_app_name

        with pytest.raises(ValueError, match="Cannot derive app name"):
            _to_fal_app_name("")

    # --- Real-world examples ---

    def test_realistic_model_names(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("StableDiffusion") == "stable-diffusion"
        assert _to_fal_app_name("TextToImage") == "text-to-image"
        assert _to_fal_app_name("ImageUpscaler") == "image-upscaler"
        assert _to_fal_app_name("VideoGenerator") == "video-generator"

    def test_realistic_snake_case_names(self):
        from fal.app import _to_fal_app_name

        assert _to_fal_app_name("stable_diffusion") == "stable-diffusion"
        assert _to_fal_app_name("text_to_image") == "text-to-image"
        assert _to_fal_app_name("flux_schnell") == "flux-schnell"


def test_app_classvars_propagate_to_host_kwargs():
    class VarsApp(App):
        request_timeout = 11
        startup_timeout = 22
        min_concurrency = 2
        max_concurrency = 3
        concurrency_buffer = 4
        concurrency_buffer_perc = 50
        scaling_delay = 33
        max_multiplexing = 7
        kind = "container"
        image = ContainerImage.from_dockerfile_str(
            "FROM python:3.10-slim",
        )
        skip_retry_conditions = ["timeout", "connection_error"]

    hk = VarsApp.host_kwargs
    assert hk["request_timeout"] == 11
    assert hk["startup_timeout"] == 22
    assert hk["min_concurrency"] == 2
    assert hk["max_concurrency"] == 3
    assert hk["concurrency_buffer"] == 4
    assert hk["concurrency_buffer_perc"] == 50
    assert hk["scaling_delay"] == 33
    assert hk["max_multiplexing"] == 7
    assert hk["kind"] == "container"
    assert isinstance(hk["image"], ContainerImage)
    assert hk["skip_retry_conditions"] == ["timeout", "connection_error"]


def test_app_files_classvars_propagate_to_host_kwargs():
    class VarsApp(App):
        request_timeout = 11
        startup_timeout = 22
        app_files = ["a.py", "b.py"]
        app_files_ignore = [r"\\.venv/"]
        app_files_context_dir = "."
        min_concurrency = 2
        max_concurrency = 3
        concurrency_buffer = 4
        concurrency_buffer_perc = 50
        scaling_delay = 33
        max_multiplexing = 7

    hk = VarsApp.host_kwargs
    assert hk["request_timeout"] == 11
    assert hk["startup_timeout"] == 22
    assert hk["app_files"] == ["a.py", "b.py"]
    assert hk["app_files_ignore"] == [r"\\.venv/"]
    assert hk["app_files_context_dir"] == "."
    assert hk["min_concurrency"] == 2
    assert hk["max_concurrency"] == 3
    assert hk["concurrency_buffer"] == 4
    assert hk["concurrency_buffer_perc"] == 50
    assert hk["scaling_delay"] == 33
    assert hk["max_multiplexing"] == 7


def test_app_kwargs_merge_into_host_kwargs_and_override_defaults():
    class KwargsApp(App, keep_alive=123, resolver="pip", _scheduler="nomad"):
        pass

    hk = KwargsApp.host_kwargs
    assert hk["keep_alive"] == 123
    assert hk["resolver"] == "pip"
    assert hk["_scheduler"] == "nomad"
    # default keys should still exist
    assert "_scheduler_options" in hk


def test_non_host_classvars_do_not_leak_into_host_kwargs():
    class LeakCheckApp(App):
        requirements = ["fastapi"]
        local_python_modules = ["."]
        machine_type = "M"
        num_gpus = 1
        app_auth = "private"

    hk = LeakCheckApp.host_kwargs
    assert "requirements" not in hk
    assert "local_python_modules" not in hk
    assert "machine_type" not in hk
    assert "num_gpus" not in hk
    assert "app_auth" not in hk


@pytest.mark.asyncio
async def test_runner_state_lifecycle_complete():
    """Test that FAL_RUNNER_STATE transitions through all phases correctly"""
    states = []

    class StateCheckApp(App):
        def setup(self):
            states.append(("setup", os.getenv("FAL_RUNNER_STATE")))

        def teardown(self):
            states.append(("teardown", os.getenv("FAL_RUNNER_STATE")))

    app = StateCheckApp(_allow_init=True)

    # Create a mock FastAPI app
    import fastapi

    fastapi_app = fastapi.FastAPI()

    async with app.lifespan(fastapi_app):
        states.append(("running", os.getenv("FAL_RUNNER_STATE")))

    # Verify the full lifecycle
    assert len(states) == 3
    assert states[0] == ("setup", "SETUP")
    assert states[1] == ("running", "RUNNING")
    assert states[2] == ("teardown", "TERMINATING")


def test_function_decorator_rejects_app_files_with_container_kind():
    """Test that app_files cannot be used with kind='container'."""
    image = ContainerImage.from_dockerfile_str("FROM python:3.11-slim")

    error_message = "app_files is not supported for container apps"
    with pytest.raises(ValueError, match=error_message):

        @fal.function("container", app_files=["a.py"])
        def container_because_kind_is_container():
            pass

    with pytest.raises(ValueError, match=error_message):

        @fal.function(image=image, app_files=["a.py"])
        def container_because_image_is_provided():
            pass


def test_app_classvars_propagate_to_host_kwargs_when_overriding_hidden_defaults():
    class VarsApp(App):
        _scheduler = "kubernetes"
        _scheduler_options = {
            "storage_region": "us-west",
        }
        keep_alive = 30
        resolver = "pip"
        _app_var = "example"

    hk = VarsApp.host_kwargs
    assert hk["_scheduler"] == "kubernetes"
    assert hk["_scheduler_options"] == {
        "storage_region": "us-west",
    }
    assert hk["keep_alive"] == 30
    assert hk["resolver"] == "pip"
    assert "_app_var" not in hk


def test_app_is_picklable_with_request_context():
    app = PickleApp(_allow_init=True)
    app._current_request_context = ContextVar(  # type: ignore[assignment]
        "_current_request_context"
    )

    payload = pickle.dumps(app)
    loaded = pickle.loads(payload)

    assert loaded._current_request_context is None



import pytest

import fal
from fal.api import IsolatedFunction, Options
from fal.utils import _find_target


class DummyHost:
    pass


def make_isolated_function(name: str = "func"):
    # Name is used only to label the symbol in the module dict
    return name, IsolatedFunction(
        host=DummyHost(), raw_func=lambda: None, options=Options()
    )


def test_find_target_by_name_missing():
    module = {}
    with pytest.raises(Exception) as exc:
        _find_target(module, "missing")
    assert "Function 'missing' not found in module" in str(exc.value)


def test_find_target_by_name_app_returns_class_and_metadata():
    class MyApp(fal.App):
        app_name = "my_app"
        app_auth = "public"

    module = {"MyApp": MyApp}
    target, app_name, app_auth, class_name = _find_target(module, "MyApp")

    assert target is MyApp
    assert app_name == MyApp.app_name
    assert app_auth == MyApp.app_auth
    assert class_name == "MyApp"


def test_find_target_by_name_isolated_function_returns_function_and_name():
    name, iso = make_isolated_function("the_func")
    module = {name: iso}

    target, returned_name, auth, class_name = _find_target(module, name)

    assert target is iso
    assert returned_name == name
    assert auth is None
    assert class_name == name


def test_find_target_by_name_invalid_type_raises():
    module = {"not_valid": 123}
    with pytest.raises(Exception) as exc:
        _find_target(module, "not_valid")
    assert "is not a fal.App or a fal.function" in str(exc.value)


def test_find_target_single_app_without_name():
    class OnlyApp(fal.App):
        app_name = "only_app"
        app_auth = "private"

    module = {"OnlyApp": OnlyApp}
    target, app_name, app_auth, class_name = _find_target(module)

    assert target is OnlyApp
    assert app_name == OnlyApp.app_name
    assert app_auth == OnlyApp.app_auth
    assert class_name == "OnlyApp"


def test_find_target_multiple_apps_without_name_raises():
    class AppOne(fal.App):
        pass

    class AppTwo(fal.App):
        pass

    module = {"AppOne": AppOne, "AppTwo": AppTwo}
    with pytest.raises(Exception) as exc:
        _find_target(module)
    assert "Multiple fal.Apps found in the module" in str(exc.value)


def test_find_target_no_apps_no_functions_raises():
    module = {"something": object()}
    with pytest.raises(Exception) as exc:
        _find_target(module)
    assert "No fal.App or fal.function found in the module" in str(exc.value)


def test_find_target_multiple_functions_without_name_raises():
    name1, iso1 = make_isolated_function("f1")
    name2, iso2 = make_isolated_function("f2")

    module = {name1: iso1, name2: iso2}
    with pytest.raises(Exception) as exc:
        _find_target(module)
    assert "Multiple fal.functions found in the module" in str(exc.value)


def test_find_target_single_function_without_name():
    name, iso = make_isolated_function("single")
    module = {name: iso}

    target, function_name, auth, class_name = _find_target(module)

    assert target is iso
    assert function_name == name
    assert auth is None
    assert class_name == name



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from fal.sdk import AuthModeLiteral, DeploymentStrategyLiteral

if TYPE_CHECKING:
    from .client import SyncServerlessClient


import json
from collections import namedtuple
from typing import TYPE_CHECKING, Tuple, Union, cast

if TYPE_CHECKING:
    from openapi_fal_rest.client import Client

User = namedtuple("User", ["user_id", "username"])


@dataclass
class DeploymentResult:
    revision: str
    app_name: str
    urls: dict[str, dict[str, str]]
    log_url: str
    auth_mode: str


def _remove_http_and_port_from_url(url):
    # Remove http://
    if "http://" in url:
        url = url.replace("http://", "")

    # Remove https://
    if "https://" in url:
        url = url.replace("https://", "")

    # Remove port information
    url_parts = url.split(":")
    if len(url_parts) > 1:
        url = url_parts[0]

    return url


def _get_user(client: Client) -> User:
    from http import HTTPStatus

    import openapi_fal_rest.api.users.get_current_user as get_current_user

    from fal.api import FalServerlessError

    try:
        user_details_response = get_current_user.sync_detailed(
            client=client,
        )
    except Exception as e:
        raise FalServerlessError(f"Error fetching user details: {str(e)}")

    if user_details_response.status_code != HTTPStatus.OK:
        try:
            content = json.loads(user_details_response.content.decode("utf8"))
        except Exception:
            raise FalServerlessError(
                f"Error fetching user details: {user_details_response}"
            )
        else:
            raise FalServerlessError(content["detail"])
    try:
        full_user_id = user_details_response.parsed.user_id
        _provider, _, user_id = full_user_id.partition("|")
        if not user_id:
            user_id = full_user_id

        return User(user_id=user_id, username=user_details_response.parsed.nickname)
    except Exception as e:
        raise FalServerlessError(f"Could not parse the user data: {e}")


def _deploy_from_reference(
    client: SyncServerlessClient,
    app_ref: Tuple[Optional[Union[Path, str]], ...],
    app_name: str,
    auth: Optional[AuthModeLiteral],
    strategy: Optional[DeploymentStrategyLiteral],
    scale: bool,
    force_env_build: bool,
    environment_name: Optional[str] = None,
) -> DeploymentResult:
    from fal.api import FalServerlessError
    from fal.utils import load_function_from

    file_path, func_name = app_ref
    if file_path is None:
        # Try to find a python file in the current directory
        options = list(Path(".").glob("*.py"))
        if len(options) == 0:
            raise FalServerlessError("No python files found in the current directory")
        elif len(options) > 1:
            raise FalServerlessError(
                "Multiple python files found in the current directory. "
                "Please specify the file path of the app you want to deploy."
            )

        [file_path] = options
        file_path = str(file_path)  # type: ignore

    host = client._create_host(
        local_file_path=str(file_path),
        environment_name=environment_name,
    )
    loaded = load_function_from(
        host,
        file_path,  # type: ignore
        func_name,  # type: ignore
        force_env_build=force_env_build,
    )
    isolated_function = loaded.function
    app_name = app_name or loaded.app_name  # type: ignore
    app_auth = auth or loaded.app_auth
    strategy = strategy or "rolling"

    # Get the actual function/class name that was loaded
    display_name = func_name or loaded.class_name

    # Show what app name will be used
    from fal.console import console

    console.print(f"Deploying '{display_name}' as app '{app_name}'", style="bold")
    console.print("")

    result = host.register(
        func=isolated_function.func,
        options=isolated_function.options,
        application_name=app_name,
        application_auth_mode=app_auth,  # type: ignore
        source_code=loaded.source_code,
        metadata=isolated_function.options.host.get("metadata", {}),
        deployment_strategy=strategy,
        scale=scale,
        environment_name=environment_name,
    )

    if not result or not result.result:
        raise FalServerlessError(
            "Deployment failed: The server did not confirm the deployment. "
            "This may indicate a network issue or server error. "
            "Please try again."
        )
    if not result.service_urls:
        raise FalServerlessError(
            "Deployment failed: Could not generate app endpoints. "
            "The app name may be invalid - try using --app-name with a simple "
            "kebab-case name (e.g., --app-name my-app)."
        )

    urls: dict[str, dict[str, str]] = {
        "playground": {},
        "sync": {},
        "async": {},
    }
    for endpoint in loaded.endpoints:
        urls["playground"][endpoint] = f"{result.service_urls.playground}{endpoint}"
        urls["sync"][endpoint] = f"{result.service_urls.run}{endpoint}"
        urls["async"][endpoint] = f"{result.service_urls.queue}{endpoint}"

    return DeploymentResult(
        revision=result.result.application_id,
        app_name=app_name,
        urls=urls,
        log_url=result.service_urls.log,
        auth_mode=app_auth or "private",
    )


def deploy(
    client: SyncServerlessClient,
    app_ref: str | tuple[str, str] | None = None,
    *,
    app_name: str | None = None,
    auth: AuthModeLiteral | None = None,
    strategy: DeploymentStrategyLiteral = "rolling",
    reset_scale: bool = False,
    force_env_build: bool = False,
    environment_name: str | None = None,
) -> DeploymentResult:
    from fal.cli._utils import get_app_data_from_toml, is_app_name
    from fal.cli.parser import RefAction

    if isinstance(app_ref, tuple):
        app_ref_tuple = app_ref
    elif app_ref:
        app_ref_tuple = RefAction.split_ref(app_ref)
    else:
        raise ValueError("Invalid app reference")

    # my-app
    if is_app_name(app_ref_tuple):
        # we do not allow --app-name and --auth to be used with app name
        if app_name or auth:
            raise ValueError("Cannot use --app-name or --auth with app name reference.")

        app_name = app_ref_tuple[0]
        app_ref, app_auth, app_strategy, app_scale_settings, _ = get_app_data_from_toml(
            app_name
        )

        file_path, func_name = RefAction.split_ref(app_ref)

    # path/to/myfile.py::MyApp
    else:
        file_path, func_name = app_ref_tuple
        app_name = cast(str, app_name)
        # default to be set in the backend
        app_auth = cast(Optional[AuthModeLiteral], auth)
        # default comes from the CLI
        app_strategy = cast(DeploymentStrategyLiteral, strategy)
        app_scale_settings = cast(bool, reset_scale)
        file_path = str(Path(file_path).absolute())

    return _deploy_from_reference(
        client,
        (file_path, func_name),
        app_name,  # type: ignore
        app_auth,
        strategy=app_strategy,
        scale=app_scale_settings,
        force_env_build=force_env_build,
        environment_name=environment_name,
    )



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.workflow_contents_metadata import WorkflowContentsMetadata
    from ..models.workflow_contents_nodes import WorkflowContentsNodes
    from ..models.workflow_contents_output import WorkflowContentsOutput
    from ..models.workflow_schema import WorkflowSchema


T = TypeVar("T", bound="WorkflowContents")


@attr.s(auto_attribs=True)
class WorkflowContents:
    """
    Attributes:
        name (str):
        nodes (WorkflowContentsNodes):
        output (WorkflowContentsOutput):
        schema (WorkflowSchema):
        version (str):
        metadata (Union[Unset, WorkflowContentsMetadata]):
    """

    name: str
    nodes: "WorkflowContentsNodes"
    output: "WorkflowContentsOutput"
    schema: "WorkflowSchema"
    version: str
    metadata: Union[Unset, "WorkflowContentsMetadata"] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        nodes = self.nodes.to_dict()

        output = self.output.to_dict()

        schema = self.schema.to_dict()

        version = self.version
        metadata: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.metadata, Unset):
            metadata = self.metadata.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "nodes": nodes,
                "output": output,
                "schema": schema,
                "version": version,
            }
        )
        if metadata is not UNSET:
            field_dict["metadata"] = metadata

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.workflow_contents_metadata import WorkflowContentsMetadata
        from ..models.workflow_contents_nodes import WorkflowContentsNodes
        from ..models.workflow_contents_output import WorkflowContentsOutput
        from ..models.workflow_schema import WorkflowSchema

        d = src_dict.copy()
        name = d.pop("name")

        nodes = WorkflowContentsNodes.from_dict(d.pop("nodes"))

        output = WorkflowContentsOutput.from_dict(d.pop("output"))

        schema = WorkflowSchema.from_dict(d.pop("schema"))

        version = d.pop("version")

        _metadata = d.pop("metadata", UNSET)
        metadata: Union[Unset, WorkflowContentsMetadata]
        if isinstance(_metadata, Unset):
            metadata = UNSET
        else:
            metadata = WorkflowContentsMetadata.from_dict(_metadata)

        workflow_contents = cls(
            name=name,
            nodes=nodes,
            output=output,
            schema=schema,
            version=version,
            metadata=metadata,
        )

        workflow_contents.additional_properties = d
        return workflow_contents

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import json
import os
import re
import shlex
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Literal, Optional

Builder = Literal["depot", "service", "worker"]
BUILDERS = {"depot", "service", "worker"}
DEFAULT_COMPRESSION: str = "gzip"
DEFAULT_FORCE_COMPRESSION: bool = False

# Default patterns to ignore
DEFAULT_DOCKERIGNORE_PATTERNS = [
    ".git",
    ".gitignore",
    ".dockerignore",
    "Dockerfile",
    "Dockerfile.*",
    "*.md",
    "__pycache__",
    "*.pyc",
    ".DS_Store",
    ".venv",
    "venv",
    ".env",
    "node_modules",
]


@dataclass
class ContainerImage:
    """ContainerImage represents a Docker image that can be built
    from a Dockerfile.
    """

    dockerfile_str: str
    build_args: Dict[str, str] = field(default_factory=dict)
    registries: Dict[str, Dict[str, str]] = field(default_factory=dict)
    builder: Optional[Builder] = field(default=None)
    compression: str = DEFAULT_COMPRESSION
    force_compression: bool = DEFAULT_FORCE_COMPRESSION
    secrets: Dict[str, str] = field(default_factory=dict)
    # Build context directory
    context_dir: os.PathLike = field(default=Path.cwd())
    dockerignore: Optional[List[str]] = field(default=None)
    dockerignore_path: Optional[os.PathLike] = field(default=None)

    def __post_init__(self) -> None:
        # Validate dockerfile first
        if not self.dockerfile_str or not self.dockerfile_str.strip():
            raise ValueError(
                "Invalid dockerfile: Dockerfile is required.\n"
                "Either use ContainerImage.from_dockerfile_str() or "
                "ContainerImage.from_dockerfile() to create a container image."
            )

        # Initialize dockerignore
        self._dockerignore = DockerignoreHandler(
            context_dir=self.context_dir,
            dockerignore=self.dockerignore,
            dockerignore_path=self.dockerignore_path,
        ).get_regex_patterns()

        if self.registries:
            for registry in self.registries.values():
                keys = registry.keys()
                if "username" not in keys or "password" not in keys:
                    raise ValueError(
                        "Username and password are required for each registry"
                    )

        if self.builder and self.builder not in BUILDERS:
            raise ValueError(
                f"Invalid builder: {self.builder}, must be one of {BUILDERS}"
            )

    @classmethod
    def from_dockerfile_str(cls, text: str, **kwargs) -> "ContainerImage":
        return cls(dockerfile_str=text, **kwargs)

    @classmethod
    def from_dockerfile(cls, path: str, **kwargs) -> "ContainerImage":
        with open(path) as fobj:
            return cls.from_dockerfile_str(fobj.read(), **kwargs)

    def to_dict(self) -> dict:
        return {
            "dockerfile_str": self.dockerfile_str,
            "build_args": self.build_args,
            "registries": self.registries,
            "builder": self.builder,
            "compression": self.compression,
            "force_compression": self.force_compression,
            "secrets": self.secrets,
            "docker_context_dir": str(self.context_dir),
            "docker_files_list": self.get_copy_add_sources(),
            "docker_ignore": self._dockerignore,
        }

    def get_copy_add_sources(self) -> List[str]:
        """
        Get list of src paths/patterns from COPY/ADD commands. This method only
        parses the Dockerfile - it doesn't access the filesystem.

        Returns:
            List of src paths (e.g., ["src/", "requirements.txt", "*.py"]) that can be
            passed to FileSync.sync_files(). Returns empty list if no COPY/ADD commands
            found.
        """
        return DockerfileParser(content=self.dockerfile_str).parse_copy_add_sources()

    def add_dockerignore(
        self,
        patterns: Optional[List[str]] = None,
        path: Optional[os.PathLike] = None,
    ) -> None:
        """Add or update dockerignore patterns.

        Sets the internal dockerignore patterns using gitignore-style matching.
        You can provide either a list of patterns or a path to a .dockerignore file.

        Args:
            patterns: List of gitignore-style patterns
            path: Path to a .dockerignore file

        Raises:
            ValueError: If both patterns and path are provided, or neither
        """
        if patterns is not None and path is not None:
            raise ValueError(
                "Cannot specify both 'patterns' and 'path'. Please provide only one."
            )

        if patterns is None and path is None:
            raise ValueError(
                "Must specify either 'patterns' or 'path'. Please provide one of them."
            )

        if path is not None:
            # Load patterns from file
            handler = DockerignoreHandler(dockerignore_path=path)
        else:
            # patterns is guaranteed to be not None here due to validation above
            handler = DockerignoreHandler(dockerignore=patterns)

        # Convert gitignore patterns to regex
        self._dockerignore = handler.get_regex_patterns()


@dataclass
class DockerignoreHandler:
    context_dir: Optional[os.PathLike] = None
    dockerignore: Optional[List[str]] = None
    dockerignore_path: Optional[os.PathLike] = None

    def _read_dockerignore_file(self, path: Path) -> List[str]:
        with open(path) as f:
            lines = f.read().splitlines()

        # Filter out empty lines and comments
        return [
            stripped
            for line in lines
            if (stripped := line.strip()) and not stripped.startswith("#")
        ]

    def get_patterns(self) -> List[str]:
        """
        Get list of ignore patterns.

        Priority (highest to lowest):
        1. Explicit dockerignore list
        2. Explicit path to the .dockerignore file
        3. .dockerignore file in the context directory
        4. Default ignore patterns

        Returns:
            List of ignore patterns
        """
        # Explicit ignore list
        if self.dockerignore is not None:
            return self.dockerignore

        # Explicit .dockerignore path
        if self.dockerignore_path is not None:
            dockerignore_path = Path(self.dockerignore_path)
            if dockerignore_path.is_file():
                return self._read_dockerignore_file(dockerignore_path)
            else:
                raise FileNotFoundError(
                    f"Specified .dockerignore file not found: {self.dockerignore_path}"
                )

        # .dockerignore in context_dir
        if self.context_dir is not None:
            context_path = Path(self.context_dir)
            dockerignore_path = context_path / ".dockerignore"
            if dockerignore_path.is_file():
                return self._read_dockerignore_file(dockerignore_path)

        # Fallback to defaults
        return DEFAULT_DOCKERIGNORE_PATTERNS

    def get_regex_patterns(self) -> List[str]:
        from pathspec.patterns.gitwildmatch import GitWildMatchPattern

        patterns = self.get_patterns()
        regex_patterns = []
        for pattern in patterns:
            # Convert ignore patterns to regex, this way we can use `re` at runtime.
            regex, _ = GitWildMatchPattern.pattern_to_regex(pattern)
            if regex:
                regex_patterns.append(regex)
        return regex_patterns


@dataclass
class DockerfileParser:
    content: str
    normalized_content: str = field(init=False)

    def __post_init__(self) -> None:
        self.normalized_content = re.sub(r"\\\n\s*", " ", self.content)

    def parse_copy_add_sources(self) -> List[str]:
        """
        Parse COPY and ADD commands to extract source paths.
            - Skips COPY --from=... (multi-stage builds)
            - Skips ADD with URLs (http://, https://)
            - Normalizes absolute paths by stripping leading slash (Docker treats
              them as relative to the build context)
            - Handles both shell form and JSON form

        Returns:
            List of source paths/patterns referenced in COPY/ADD commands.
        """
        sources: List[str] = []

        # Regex to match COPY or ADD instructions
        instruction_pattern = re.compile(
            r"^(?P<instruction>COPY|ADD)\s+(?P<rest>.+?)(?:\s*\\)?$",
            re.MULTILINE | re.IGNORECASE,
        )

        for match in instruction_pattern.finditer(self.normalized_content):
            instruction = match.group("instruction").upper()
            rest = match.group("rest").strip()

            if not self._is_valid_source(rest, instruction):
                continue

            src_paths = self._parse_instruction_args(rest, instruction)
            sources.extend(src_paths)

        return sources

    def _is_valid_source(self, args: str, instruction: str) -> bool:
        # Skip COPY --from=... (multi-stage builds)
        if instruction == "COPY" and re.match(r"--from=", args, re.IGNORECASE):
            return False

        # Skip ADD with URLs (e.g., ADD https://example.com/file.tar.gz)
        # Use re.search to handle cases where flags precede the URL
        if instruction == "ADD" and re.search(r"https?://", args, re.IGNORECASE):
            return False

        return True

    def _parse_instruction_args(self, args_str: str, instruction: str) -> List[str]:
        """
        Parse arguments from COPY/ADD instruction.

        Args:
            args_str: The arguments string after the instruction
            instruction: The instruction type (COPY or ADD)

        Returns:
            List of valid source paths
        """
        src_paths: List[str] = []

        try:
            # Remove flags from the args string first
            # Matches: --flag or --flag=value or --flag="value"
            r = r'--\w+(?:=[^\s\]]+|="[^"]*")?'
            args_str_clean = re.sub(r, "", args_str).strip()

            if not args_str_clean:
                return []

            # Try JSON form first: ["src1", "src2", "dest"]
            if args_str_clean.endswith("]"):
                args = json.loads(args_str_clean)
            else:
                args = shlex.split(args_str_clean)

            if len(args) >= 2:
                # All but last are sources
                src_paths = args[:-1]

            # We normalize to relative paths (strip leading "/") because:
            # - They're semantically identical in Docker
            # - FileSync requires relative paths to context_dir
            return [src_path.lstrip("/") for src_path in src_paths]

        except (AttributeError, json.JSONDecodeError, TypeError, ValueError):
            from fal.console import console

            console.print(
                f"[yellow][WARNING][/yellow] Failed to parse instruction arguments: "
                f"{args_str}. Skipping this instruction. Please check the Dockerfile "
                "syntax and try again."
            )
            return []



from functools import wraps

from pydantic import Field

from fal.toolkit.file.file import IS_PYDANTIC_V2, File


@wraps(Field)
def AudioField(*args, **kwargs):
    ui = kwargs.get("ui", {})
    ui.setdefault("field", "audio")
    kwargs["ui"] = ui
    return Field(*args, **kwargs)


class Audio(File):
    if IS_PYDANTIC_V2:
        model_config = {"json_schema_extra": {"ui": {"field": "audio"}}}
    else:

        class Config:
            @staticmethod
            def schema_extra(schema, model_type):
                schema.setdefault("ui", {})["field"] = "audio"



import posixpath
import uuid

from fal.files import FalFileSystem


def test_fal_fs(tmp_path):
    (tmp_path / "myfile").write_text("myfile")
    (tmp_path / "mydir").mkdir()
    (tmp_path / "mydir" / "myfile1").write_text("myfile1")
    (tmp_path / "mydir" / "myfile2").write_text("myfile2")

    (tmp_path / "downloaded").mkdir()

    remote_temp_dir = f"/data/tmp/{uuid.uuid4()}"

    fs = FalFileSystem()

    assert fs.exists("/")
    assert fs.isdir("/")
    assert fs.exists("/data")
    assert fs.isdir("/data")
    assert fs.exists(".")
    assert fs.isdir(".")

    fs.put(str(tmp_path / "myfile"), remote_temp_dir + "/myfile")
    assert fs.exists(remote_temp_dir + "/myfile")
    assert fs.isdir(remote_temp_dir)
    assert fs.isfile(remote_temp_dir + "/myfile")
    assert not fs.isdir(remote_temp_dir + "/myfile")
    fs.get(remote_temp_dir + "/myfile", str(tmp_path / "downloaded" / "myfile"))
    assert (tmp_path / "downloaded" / "myfile").read_text() == "myfile"

    fs.put(str(tmp_path / "mydir"), remote_temp_dir + "/mydir", recursive=True)
    assert fs.exists(remote_temp_dir + "/mydir")
    assert fs.isdir(remote_temp_dir + "/mydir")
    assert fs.isfile(remote_temp_dir + "/mydir/myfile1")
    assert fs.isfile(remote_temp_dir + "/mydir/myfile2")
    fs.get(
        remote_temp_dir + "/mydir",
        str(tmp_path / "downloaded" / "mydir"),
        recursive=True,
    )
    assert (tmp_path / "downloaded" / "mydir" / "myfile1").read_text() == "myfile1"
    assert (tmp_path / "downloaded" / "mydir" / "myfile2").read_text() == "myfile2"

    assert fs.ls(remote_temp_dir, detail=False) == [
        remote_temp_dir + "/mydir",
        remote_temp_dir + "/myfile",
    ]
    assert fs.ls(posixpath.relpath(remote_temp_dir, "/data"), detail=False) == [
        remote_temp_dir + "/mydir",
        remote_temp_dir + "/myfile",
    ]

    fs.rm(remote_temp_dir + "/myfile")
    assert not fs.exists(remote_temp_dir + "/myfile")
    fs.rm(remote_temp_dir + "/mydir", recursive=True)
    assert not fs.exists(remote_temp_dir + "/mydir")



from __future__ import annotations

import os
import traceback
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import torch


def optimize(
    module: torch.nn.Module, *, optimization_config: dict[str, Any] | None = None
) -> torch.nn.Module:
    """Optimize the given torch module with dynamic compilation and
    quantization techniques. Only applicable under fal's cloud environment.

    Warning: This function is experimental and may not work as expected.
    """
    import runpy

    import torch.nn as nn

    if not isinstance(module, nn.Module):
        raise TypeError(f"Expected a torch.nn.Module, got {type(module)}.")

    optimizer_path = os.environ.get("FAL_SPATIAL_OPTIMIZER", None)
    if optimizer_path is None:
        print(
            "[WARNING] FAL_SPATIAL_OPTIMIZER is not set, falling back"
            "to default torch execution"
        )
        return module

    try:
        spatial_optimizer = runpy.run_path(optimizer_path, run_name="spatial_optimizer")

        return spatial_optimizer["optimize"](
            module,
            optimization_config=optimization_config,
        )
    except Exception:
        print(
            "[WARNING] Failed to optimize module, falling back "
            "to default torch execution."
        )
        traceback.print_exc()
        return module



from .api import *  # noqa: F403
from .client import (  # noqa: F401
    AppsNamespace,
    KeysNamespace,
    RunnersNamespace,
    SecretsNamespace,
    SyncServerlessClient,
)

__all__ = [
    "SyncServerlessClient",
    "AppsNamespace",
    "RunnersNamespace",
    "KeysNamespace",
    "SecretsNamespace",
]



import concurrent.futures
import hashlib
import os
import re
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path, PurePosixPath
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Pattern, Tuple

import httpx
from rich.tree import Tree

import fal.flags as flags
from fal._version import version_tuple
from fal.console import console
from fal.console.icons import CROSS_ICON
from fal.exceptions import (
    AppFileUploadException,
    FalServerlessException,
    FileTooLargeError,
)
from fal.upload import (
    MULTIPART_CHUNK_SIZE,
    AppFileMultipartUpload,
)

if TYPE_CHECKING:
    from fal.api.api import Options

USER_AGENT = f"fal-sdk/{'.'.join(map(str, version_tuple))} (python)"
FILE_SIZE_LIMIT = 1024 * 1024 * 1024  # 1GB
DEFAULT_CONCURRENCY_UPLOADS = 10


@dataclass
class FileSyncOptions:
    files_list: List[str]
    files_ignore: List[Pattern]
    files_context_dir: Optional[str]

    @classmethod
    def from_options(cls, options: "Options") -> "FileSyncOptions":
        # Container files
        if options.environment.get("kind") == "container":
            image_dict: Dict[str, Any] = options.environment.get("image", {})
            files_list = image_dict.get("docker_files_list", [])
            files_ignore = image_dict.get("docker_ignore", [])
            files_context_dir = image_dict.get("docker_context_dir")

        else:  # App files
            files_list = options.host.get("app_files", [])
            files_ignore = options.host.get("app_files_ignore", [])
            files_context_dir = options.host.get("app_files_context_dir")

        # Compile regex patterns (both containers and app_files use regex now)
        files_ignore = [re.compile(pattern) for pattern in files_ignore]
        return cls(
            files_list=files_list,
            files_ignore=files_ignore,
            files_context_dir=files_context_dir,
        )


def print_path_tree(file_paths):
    tree = Tree("/app")

    nodes = {"": tree}

    for file_path in sorted(file_paths):
        parts = Path(file_path).parts

        for i, part in enumerate(parts):
            current_path = str(Path(*parts[: i + 1]))

            if current_path not in nodes:
                parent_path = str(Path(*parts[:i])) if i > 0 else ""
                parent_node = nodes[parent_path]

                nodes[current_path] = parent_node.add(f"{part}")

    console.print(tree)


def sanitize_relative_path(rel_path: str, original_path: Path) -> str:
    pure_path = PurePosixPath(rel_path)

    # Block files that are absolute or contain parent directory references
    if pure_path.is_absolute():
        raise FalServerlessException(f"Absolute Path is not allowed: {rel_path}")
    if ".." in pure_path.parts or "." in pure_path.parts:
        raise FalServerlessException(
            "Parent directory reference is not allowed: "
            + f"{rel_path} for {original_path}\n"
            + "If you didn't mean to sync this file, please ignore it using "
            + "`app_files_ignore`."
        )

    return pure_path.as_posix()


# This is the same function we use in the backend to compute file hashes
# It is important that this is the same
def compute_hash(file_path: Path, mode: int) -> str:
    file_hash = hashlib.sha256()
    with file_path.open("rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            file_hash.update(chunk)

    # Include metadata in hash
    metadata_string = f"{mode}"
    file_hash.update(metadata_string.encode("utf-8"))

    return file_hash.hexdigest()


def normalize_path(
    path_str: str, base_path_str: str, files_context_dir: Optional[str] = None
) -> Tuple[str, str]:
    path = Path(path_str)
    base_path = Path(base_path_str).resolve()

    if base_path.is_dir():
        script_dir = base_path
    else:
        script_dir = base_path.parent

    if files_context_dir:
        context_path = Path(files_context_dir)
        if context_path.is_absolute():
            script_dir = context_path.resolve()
        else:
            script_dir = (script_dir / context_path).resolve()

    absolute_path = (
        path.resolve() if path.is_absolute() else (script_dir / path).resolve()
    )

    try:
        relative_path = os.path.relpath(absolute_path, script_dir)
        relative_path = sanitize_relative_path(relative_path, path)
    except ValueError:
        raise ValueError(f"Invalid relative path: {absolute_path}")

    return absolute_path.as_posix(), relative_path


@dataclass
class FileMetadata:
    size: int
    mtime: float
    mode: int
    hash: str
    relative_path: str
    absolute_path: str

    @classmethod
    def from_path(
        cls, file_path: Path, *, relative: str, absolute: str
    ) -> "FileMetadata":
        stat = file_path.stat()
        # Limit allowed individual file size
        if stat.st_size > FILE_SIZE_LIMIT:
            raise FileTooLargeError(
                message=f"{file_path} is larger than {FILE_SIZE_LIMIT} bytes."
            )

        file_hash = compute_hash(file_path, stat.st_mode)
        return FileMetadata(
            size=stat.st_size,
            mtime=stat.st_mtime,
            mode=stat.st_mode,
            hash=file_hash,
            relative_path=relative,
            absolute_path=absolute,
        )

    def to_dict(self) -> Dict[str, str]:
        return {
            "size": str(self.size),
            "mtime": str(self.mtime),
            "mode": str(self.mode),
            "hash": self.hash,
        }


class FileSync:
    def __init__(self, local_file_path: str):
        from fal.sdk import get_default_credentials  # noqa: PLC0415

        self.creds = get_default_credentials()
        self.local_file_path = local_file_path

    @cached_property
    def _client(self) -> httpx.Client:
        from fal.flags import REST_URL

        return httpx.Client(
            base_url=REST_URL,
            headers={
                **self.creds.to_headers(),
                "User-Agent": USER_AGENT,
            },
            timeout=httpx.Timeout(
                connect=30,
                read=4 * 60,  # multipart complete can take time
                write=5 * 60,  # we could be uploading slowly
                pool=30,
            ),
        )

    def _request(self, method: str, path: str, **kwargs) -> httpx.Response:
        response = self._client.request(method, path, **kwargs)
        if response.status_code == 404:
            raise FalServerlessException("Not Found")
        elif response.status_code != 200:
            try:
                detail = response.json()["detail"]
            except Exception:
                detail = response.text
            raise FalServerlessException(detail)
        return response

    def collect_files(self, paths: List[str], files_context_dir: Optional[str] = None):
        collected_files: List[FileMetadata] = []

        for path in paths:
            abs_path_str, rel_path = normalize_path(
                path, self.local_file_path, files_context_dir
            )
            abs_path = Path(abs_path_str)
            if not abs_path.exists():
                console.print(f"{abs_path} was not found, it will be skipped")
                continue

            if abs_path.is_file():
                metadata = FileMetadata.from_path(
                    abs_path, relative=rel_path, absolute=abs_path_str
                )
                collected_files.append(metadata)

            elif abs_path.is_dir():
                # Recursively walk directory tree
                for file_path in abs_path.rglob("*"):
                    if file_path.is_file():
                        file_abs_str, file_rel_path = normalize_path(
                            str(file_path), self.local_file_path, files_context_dir
                        )
                        metadata = FileMetadata.from_path(
                            Path(file_abs_str),
                            relative=file_rel_path,
                            absolute=file_abs_str,
                        )
                        collected_files.append(metadata)

        return collected_files

    def check_hashes_on_server(self, hashes: List[str]) -> List[str]:
        try:
            response = self._request(
                "POST", "/files/missing_hashes", json={"hashes": hashes}
            )
            response.raise_for_status()
            data = response.json()
            return data
        except Exception as e:
            console.print(f"{CROSS_ICON} Failed to check hashes on server: {e}")

            return hashes

    def upload_file_multipart(
        self,
        file_path: str,
        metadata: FileMetadata,
        chunk_size: int = MULTIPART_CHUNK_SIZE,
    ) -> str:
        multipart = AppFileMultipartUpload(
            client=self._client,
            file_hash=metadata.hash,
            metadata=metadata.to_dict(),
            chunk_size=chunk_size,
        )
        try:
            return multipart.upload_file(file_path)
        except Exception as e:
            raise AppFileUploadException(str(e), metadata.relative_path)

    def _matches_patterns(self, relative_path: str, patterns: List[re.Pattern]) -> bool:
        return any(pattern.search(relative_path) for pattern in patterns)

    def sync_files(
        self,
        paths: List[str],
        chunk_size: int = MULTIPART_CHUNK_SIZE,
        max_concurrency_uploads: int = DEFAULT_CONCURRENCY_UPLOADS,
        files_ignore: List[re.Pattern] = [],
        files_context_dir: Optional[str] = None,
    ) -> Tuple[List[FileMetadata], List[AppFileUploadException]]:
        files = self.collect_files(paths, files_context_dir)

        # Filter out ignored files
        if files_ignore:
            filtered_files: List[FileMetadata] = []
            for metadata in files:
                if self._matches_patterns(metadata.relative_path, files_ignore):
                    if flags.DEBUG:
                        console.print(f"Ignoring file: {metadata.relative_path}")
                else:
                    filtered_files.append(metadata)

            # Update files list
            files = filtered_files

        # Remove duplicate files by absolute path
        unique_files: List[FileMetadata] = []
        seen_paths = set()
        seen_relative_paths = set()
        for metadata in files:
            abs_path = metadata.absolute_path
            rel_path = metadata.relative_path
            if abs_path not in seen_paths:
                seen_paths.add(abs_path)
                if rel_path in seen_relative_paths:
                    raise Exception(
                        f"Duplicate relative path '{rel_path}' found for '{abs_path}'"
                    )
                seen_relative_paths.add(rel_path)
                unique_files.append(metadata)
            else:
                if rel_path not in seen_relative_paths:
                    seen_relative_paths.add(rel_path)

        if not unique_files:
            return [], []

        hashes_to_check = list({metadata.hash for metadata in unique_files})
        missing_hashes = set(self.check_hashes_on_server(hashes_to_check))

        # Categorize based on server response
        files_to_upload: List[FileMetadata] = []
        for file in unique_files:
            if file.hash in missing_hashes:
                # No longer missing as we are uploading it
                # Removing it avoids duplicate uploads
                missing_hashes.remove(file.hash)
                files_to_upload.append(file)

        uploaded_files: List[Tuple[FileMetadata, str]] = []
        errors: List[AppFileUploadException] = []
        # Upload missing files in parallel with bounded concurrency
        if files_to_upload:
            # Embed it here to be able to pass it to the executor
            def upload_single_file(metadata: FileMetadata):
                console.print(f"Uploading file: {metadata.relative_path}")
                return self.upload_file_multipart(
                    metadata.absolute_path,
                    chunk_size=chunk_size,
                    metadata=metadata,
                )

            with concurrent.futures.ThreadPoolExecutor(
                max_workers=max_concurrency_uploads
            ) as executor:
                futures = [
                    executor.submit(upload_single_file, metadata)
                    for metadata in files_to_upload
                ]

                concurrent.futures.wait(futures)
                for metadata, future in zip(files_to_upload, futures):
                    if exc := future.exception():
                        errors.append(
                            AppFileUploadException(str(exc), metadata.relative_path)
                        )
                    else:
                        uploaded_files.append((metadata, future.result()))

        if flags.DEBUG:
            console.print("File Structure:")
            print_path_tree([m.relative_path for m in unique_files])

        return unique_files, errors

    def close(self):
        self._client.close()



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, List, Optional

from fal.sdk import AliasInfo, FalServerlessClient, RunnerInfo

if TYPE_CHECKING:
    from .client import SyncServerlessClient


def list_apps(
    client: SyncServerlessClient,
    *,
    filter: Optional[str] = None,
    environment_name: Optional[str] = None,
) -> List[AliasInfo]:
    with FalServerlessClient(client._grpc_host, client._credentials).connect() as conn:
        apps = conn.list_aliases(environment_name=environment_name)

    if filter:
        apps = [a for a in apps if filter in a.alias]
    return apps


def apps_runners(
    client: SyncServerlessClient,
    app_name: str,
    *,
    since: Optional[datetime] = None,
    state: Optional[list[str]] = None,
    environment_name: Optional[str] = None,
) -> List[RunnerInfo]:
    with FalServerlessClient(client._grpc_host, client._credentials).connect() as conn:
        alias_runners = conn.list_alias_runners(
            alias=app_name, start_time=since, environment_name=environment_name
        )

    if state and "all" not in set(state):
        states = set(state)
        alias_runners = [
            r
            for r in alias_runners
            if r.state.value.lower() in states
            or (
                "terminated" in states and r.state.value.lower() == "dead"
            )  # TODO for backwards compatibility. remove later
        ]
    return alias_runners


def scale_app(
    client: SyncServerlessClient,
    app_name: str,
    *,
    keep_alive: int | None = None,
    max_multiplexing: int | None = None,
    max_concurrency: int | None = None,
    min_concurrency: int | None = None,
    concurrency_buffer: int | None = None,
    concurrency_buffer_perc: int | None = None,
    scaling_delay: int | None = None,
    request_timeout: int | None = None,
    startup_timeout: int | None = None,
    machine_types: list[str] | None = None,
    regions: list[str] | None = None,
    environment_name: str | None = None,
) -> AliasInfo:
    with FalServerlessClient(client._grpc_host, client._credentials).connect() as conn:
        return conn.update_application(
            application_name=app_name,
            keep_alive=keep_alive,
            max_multiplexing=max_multiplexing,
            max_concurrency=max_concurrency,
            min_concurrency=min_concurrency,
            concurrency_buffer=concurrency_buffer,
            concurrency_buffer_perc=concurrency_buffer_perc,
            scaling_delay=scaling_delay,
            request_timeout=request_timeout,
            startup_timeout=startup_timeout,
            machine_types=machine_types,
            valid_regions=regions,
            environment_name=environment_name,
        )


def rollout_app(
    client: SyncServerlessClient,
    app_name: str,
    *,
    force: bool = False,
    environment_name: str | None = None,
) -> None:
    with FalServerlessClient(client._grpc_host, client._credentials).connect() as conn:
        conn.rollout_application(
            application_name=app_name,
            force=force,
            environment_name=environment_name,
        )
